/*
 * This file is part of Jstacs.
 *
 * Jstacs is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jstacs is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Jstacs.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * For more information on Jstacs, visit http://www.jstacs.de
 */

package de.jstacs.scoringFunctions.directedGraphicalModels;

import de.jstacs.NonParsableException;
import de.jstacs.Storable;
import de.jstacs.data.Alphabet;
import de.jstacs.data.Sequence;
import de.jstacs.io.XMLParser;

/**
 * Class for the parameters of a {@link BayesianNetworkScoringFunction}. Each parameter holds its current value, the symbol and position it is responsible for, 
 * the context of the current symbol. A parameter can either be free or determined by other parameters (on the same simplex).
 * 
 * @author Jan Grau
 *
 */
public class Parameter implements Storable, Cloneable {

	/**
	 * The current value of this parameter
	 */
	private double value;
	/**
	 * The symbol (out of some {@link Alphabet}) this parameter is responsible for.
	 */
	protected byte symbol;
	/**
	 * The position of <code>symbol</code> this parameter is responsible for.
	 */
	protected int position;
	/**
	 * The context of this parameter. The length <code>context.length</code> is equal to the number of parents, 
	 * <code>context[i]</code> holds the information for parent no <code>i</code>, <code>context[i][0]</code> holds the position of
	 * parent <code>i</code>, <code>context[i][1]</code>,... hold the possible configurations of parent <code>i</code>
	 * in the context. Normally, only <code>context[i][1]</code> exists, i.e. only one configuration is allowed in the context.
	 */
	protected int[][] context;
	/**
	 * The counts for this parameter. Used for determination of plug-in parameters.
	 */
	protected double count;
	/**
	 * The pseudo count for this parameter. Used for determination of plug-in parameters.
	 */
	protected double pseudoCount;
	private boolean free;

	private int index;
	private double expValue;
	private Double z;
	private Double t;
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Parameter clone() throws CloneNotSupportedException{
		Parameter clone = (Parameter) super.clone();
		for(int i=0;i<context.length;i++){
			clone.context[i] = context[i].clone();
		}
		return clone;
	}
	
	/**
	 * Creates a new parameter, that is parameter no <code>index</code> in the list of parameters of the {@link BayesianNetworkScoringFunction} and responsible for <code>symbol</code>
	 * at position <code>position</code> and pseudo count <code>pseudoCount</code>.
	 * @param index the index in the list of all parameters
	 * @param symbol the symbol this parameter is responsible for
	 * @param position the position of the symbol
	 * @param pseudoCount the pseudo count
	 * @param free indicates if this parameter is a free parameter
	 */
	public Parameter(int index, byte symbol, int position, double pseudoCount, boolean free){
		this(index, symbol,position,new int[0][2],pseudoCount,free);
	}
	
	/**
	 * Creates a new parameter, that is parameter no <code>index</code> in the list of parameters of the {@link BayesianNetworkScoringFunction} and responsible for <code>symbol</code>
	 * at position <code>position</code> having context <code>context</code> and pseudo count <code>pseudoCount</code>.
	 * @param index the index in the list of all parameters
	 * @param symbol the symbol this parameter is responsible for
	 * @param position the position of the symbol
	 * @param context the context of this parameter
	 * @param pseudoCount the pseudo count
	 * @param free indicates if this parameter is a free parameter
	 * @see Parameter#context
	 */
	public Parameter(int index,byte symbol, int position, int[][] context, double pseudoCount,boolean free){
		this.index = index;
		this.symbol = symbol;
		this.position = position;
		this.context = context;
		this.pseudoCount = pseudoCount;
		this.count = pseudoCount;
		this.value = 0;
		this.expValue = 1;
		this.free = free;
		this.z = null;
		this.t = null;
	}
	
	/**
	 * Re-creates a parameter from its XML-representation as returned by the {@link Parameter#toXML()}} method.
	 * @param representation the XML-code
	 * @throws NonParsableException is thrown if the representation could not be parsed
	 */
	public Parameter(StringBuffer representation) throws NonParsableException{
		representation = XMLParser.extractForTag(representation, "parameter");
		value = XMLParser.extractDoubleForTag(representation, "value");
		expValue = Math.exp(value);
		index = XMLParser.extractIntForTag( representation, "index" );
		pseudoCount = XMLParser.extractDoubleForTag( representation, "pseudoCount" );
		symbol = XMLParser.extractByteForTag(representation, "symbol");
		position = XMLParser.extractIntForTag(representation, "position");
		context = XMLParser.extractInt2ArrayForTag(representation, "context");
		count = XMLParser.extractDoubleForTag(representation, "count");
		free = XMLParser.extractBooleanForTag(representation, "free");
		String temp = XMLParser.extractStringForTag(representation, "z");
		z = temp.equals("null") ? null : Double.parseDouble(temp);
		temp = XMLParser.extractStringForTag(representation, "t");
		t = temp.equals("null") ? null : Double.parseDouble(temp);
	}
	
	/**
	 * Returns the pseudo count as given in the constructor.
	 * @return the pseudo count
	 * @see Parameter#Parameter(int, byte, int, double, boolean)
	 * @see Parameter#Parameter(int, byte, int, int[][], double, boolean)
	 */
	public double getPseudoCount(){
		return pseudoCount;
	}
	
	/**
	 * Resets the counts to the pseudo counts and the value to <code>0</code>.
	 */
	public void reset(){
		this.count = pseudoCount;
		this.value = 0;
		this.expValue = 1;
	}
	
	/**
	 * Returns the depth of the tree, i.e. the number of parents of this parameter.
	 * @return the depth of the tree
	 */
	public int getDepth(){
		return context.length;
	}
	
	
	/**
	 * Prints the counts and the value of this parameter to {@link System#out}.
	 */
	public void print(){
		System.out.println(symbol+" c: "+count);
		System.out.println(symbol+": "+value);
	}
	
	/**
	 * Returns true if <code>seq</code> fulfills all requirements defined in the context ({@link Parameter#context}}).
	 * @param seq the sequence
	 * @return if this parameter is responsible for <code>seq</code>
	 */
	public double doesApplyFor(Sequence seq){
		if(seq.discreteVal(position) == symbol){
			for(int i=0;i<context.length;i++){
				boolean app = false;
				for(int j=1;j<context[i].length;j++){
					if(seq.discreteVal(context[i][0]) == context[i][j]){
						app = true;
					}
				}
				if(!app){
					return 0.0;
				}
			}
			return 1.0;
		}else{
			return 0.0;
		}
	}
	
	/**
	 * Returns the current value of this parameter.
	 * @return the value
	 */
	public double getValue() {
		return value;
	}

	/**
	 * Sets the current value of this parameter.
	 * @param value the new value
	 */
	public void setValue(double value) {
		this.value = value;
		this.expValue = Math.exp(value);
	}
	
	/**
	 * Returns <code>Math.exp(getValue())</code>, which is pre-computed.
	 * @return the exp-value
	 */
	public double getExpValue(){
		return this.expValue;
	}

	
	/* (non-Javadoc)
	 * @see de.jstacs.Storable#toXML()
	 */
	public StringBuffer toXML() {
		StringBuffer buf = new StringBuffer();
		XMLParser.appendDoubleWithTags(buf, value, "value");
		XMLParser.appendByteWithTags(buf, symbol, "symbol");
		XMLParser.appendIntWithTags( buf, index, "index" );
		XMLParser.appendDoubleWithTags( buf, pseudoCount, "pseudoCount" );
		XMLParser.appendIntWithTags(buf, position, "position");
		XMLParser.appendInt2ArrayWithTags(buf, context, "context");
		XMLParser.appendDoubleWithTags(buf, count, "count");
		XMLParser.appendBooleanWithTags(buf, free, "free");
		if(z != null){
			XMLParser.appendDoubleWithTags(buf, z, "z");
		}else{
			XMLParser.appendStringWithTags(buf, null, "z");
		}
		if(t != null){
			XMLParser.appendDoubleWithTags(buf, t, "t");
		}else{
			XMLParser.appendStringWithTags(buf, "null", "t");
		}
		
		XMLParser.addTags(buf, "parameter");
		return buf;
	}
	
	/**
	 * Resets all internal normalization constants
	 */
	public void invalidateNormalizers(){
		this.z = null;
		this.t = null;
	}
	
	/**
	 * Returns the partial derivative of the normalization constant with respect to this parameter.
	 * @return the partial derivative
	 * @throws Exception thrown if no normalization constants have been pre-computed
	 */
	public double getPartialNormalizer() throws Exception{
		if(z == null || t == null){
			throw new Exception("No valid normalizers available for parameter "+index+" at position "+position+": z="+z+", t="+t+".");
		}
		return t * expValue * z;
	}

	/**
	 * Returns the current counts for this parameter.
	 * @return the counts
	 */
	public double getCounts(){
		return count;
	}
	
	/**
	 * Adds <code>count2</code> the the counts of this parameter.
	 * @param count2 the additional counts
	 */
	public void addCount(double count2) {
		this.count += count2;
	}

	/**
	 * Indicates if this parameter is free.
	 * @return if the parameter is free
	 */
	public boolean isFree() {
		return free;
	}

	/**
	 * Returns the position of this parameter as defined in the constructor.
	 * @return the position
	 * @see Parameter#Parameter(int, byte, int, double, boolean)
	 * @see Parameter#Parameter(int, byte, int, int[][], double, boolean)
	 */
	public int getPosition() {
		return position;
	}

	/**
	 * Returns the index of this parameter as defined in the constructor.
	 * @return the position
	 * @see Parameter#Parameter(int, byte, int, double, boolean)
	 * @see Parameter#Parameter(int, byte, int, int[][], double, boolean)
	 */
	public int getIndex() {
		return index;
	}

	/**
	 * Sets the part of the normalization constant of parameters before this parameter in the structure of the network.
	 * @param t the normalization constant
	 */
	void setT(Double t) {
		this.t = t;
	}

	/**
	 * Sets the part of the normalization constant of parameters after this parameter in the structure of the network.
	 * @param z the normalization constant
	 */
	void setZ(Double z) {
		this.z = z;
	}

	/**
	 * Returns the part of the normalization constant of parameters after this parameter in the structure of the network.
	 * @return the normalization constant
	 */
	public double getZ() {
		return z;
	}


}
